## mutate.m - Performs natural selection, mating and mutation on chromosomes
##           to generate next generation for analysis     
## Author:   Richard James MacPherson and Grant Ingram
#####################

# Copyright 2009-2012  Richard James MacPherson, Grant Ingram, 
# Jamie McIntosh and Michael Hilfer.  
# This file is part of PEW V2.

# PEW is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PEW is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PEW.  If not, see <http://www.gnu.org/licenses/>.

# Revision History - Jamie McIntosh
# 1.01 - Random mutation modified to assign values correctly and
#        scales them between fourier limits - perturbation scaling
#        handled in curve.m
# 1.02 - Modified scaling after mutation to work correctly with
#        amplitude values in par matrix
# 1.03 - Survival into each generation is now determined by a probability
#        density function. It is no longer selective which reduces the 
#        chance of converging to a local optimum 
# 1.04 - Survivors of each generation are copied into the next Generation
#        filestore so that repeat testing is removed and script modified to
#        skip these chromosomes in CFD analysis
# 1.05 - Reverted to 1.03 - skipping affects mutation
# 1.06 - Fittest chromosome now displayed in log file
# 1.07 - Option added to determine if elitist/stochastic method should be used
# 1.08 - Hybrid selection scheme integrated. The number of stochastic selections
#        performed is determined by hybrid_routine set in GA.m
#
#####################
# Known Issues
#
#####################

function mutate()

global iga; global harmonics; global popsize; global npar;
global par; global base_dir; global cluster_dir; global job_id;
global maxit; global cost; global Nt; global keep; global nmut;
global harmonics; global constantshi; global constantslo;
global heighthi; global heightlow; global minc; global meanc;
global hybrid_routine;

# Add chromosome label
par(:,npar+2)=1:popsize;
printf("\n popsize=%d par=%d \n", par, popsize);
# Sort data for natural selection and determine GA performance parameters
par=sortrows(par,(npar+1));   % Finding the minimum cost in each generation
minc(iga+1)=min(cost);

# Finding the mean cost in each generation
meancost=0; num=0;      
for chrom=1:1:(popsize) 
  if cost(chrom,1) ~= 1000  % Ignore chromosomes not meshed or with CFD failed
    meancost=meancost+cost(chrom,1);
    num=num+1;
  endif
endfor

meanc(iga+1)=meancost/num;

fittest = par(1,npar+2); % Saves the fittest chromosome

par(:,(npar+1:npar+2))=[]; % Clearing the cost and label from the par file

## Save updated log including cost
filename = ['Gen_' int2str(iga)]; 
save(filename, 'log', 'iga', 'par', 'fittest', 'minc', 'meanc');

## Natural Selection Routine
S=keep;				                               % Number of Survivors
prob1=flipud([1:popsize]'/sum([1:popsize])); % Weights Chromosomes
odds1=[0 cumsum(prob1(1:popsize))'];         % Probability Distribution Function
  
## Pair and mate
M=ceil((popsize-keep)/2);		              % Number of Matings
prob2=flipud([1:keep]'/sum([1:keep]));    % Weights Chromosomes
odds2=[0 cumsum(prob2(1:keep))'];     	  % Probability Distribution Function
pick1=rand(1,M);			    		            % Mate #1
pick2=rand(1,M);			    		            % Mate #2

## Selection Method
if iga >= hybrid_routine
  adaptive_limit = 0; % Trigger for elitist selection scheme
else
  adaptive_limit = 1; % Trigger for stochastic selection scheme
endif

## STOCHASTIC SELECTION SCHEME
if adaptive_limit == 1
  ## Determine the chromosomes that survive to the next generation
  ia=1;
  survivors=0;
  printf('Fittest individuals surviving to the next generation: '); % Print survival output, NOTE: this is the ranked value not chrom
  while ia<=S
    surv=rand;
    ib=2;
    while ib<=popsize+1
      if surv<=odds1(ib) && surv>odds1(ib-1)
        newpar(ia,:)=par(ib-1,:);
        reset=0;   % Match found - reset latch
        survivors(ia)=ib-1;
        if (any (survivors(1:size(survivors,2)-1) == ib-1)) == 1 && ia != 1 % Checks for duplicates
          reset=1; % Duplicate found - trigger latch
        endif 
      endif
      ib++;
    endwhile
    if reset != 1 % If latch is triggered - survival loop will repeat to remove duplicate
      ia++;
    endif
  endwhile
  printf('%d ',survivors);

  ## Sort rows so chromosome weighting correctly applied in mating 
  newpar(:,npar+1)=survivors';
  newpar=sortrows(newpar,(npar+1));
  newpar(:,(npar+1))=[]; % Clears the ranking from the matrix

## ELITIST SELECTION SCHEME
else
  newpar(:,:)=par(1:S,:);
endif

## Detemines the ma and pa indicies of the chromosomes that will mate
ic=1;
while ic<=M
  for id=2:keep+1
    if pick1(ic)<=odds2(id) && pick1(ic)>odds2(id-1)
      ma(ic)=id-1;
    endif
    if pick2(ic)<=odds2(id) && pick2(ic)>odds2(id-1)
      pa(ic)=id-1;
    endif
  endfor
  ic=ic+1;
endwhile
  
## Mating Routine using Single Point Crossover

ix=1:2:keep;                        # Index of Mate #1
xp=ceil(rand(1,M)*Nt);  	          # Crossover Point
r=rand(1,M);                        # Mixing Parameter
for ic=1:M
  xy=newpar(ma(ic),xp(ic))-newpar(pa(ic),xp(ic));  % Ma and Pa mate
  newpar(keep+ix(ic),:)=newpar(ma(ic),:);          % 1st Offspring
  newpar(keep+ix(ic)+1,:)=newpar(pa(ic),:);        % 2nd Offspring
  newpar(keep+ix(ic),xp(ic))=newpar(ma(ic),xp(ic))-r(ic).*xy;   % 1st     
  newpar(keep+ix(ic)+1,xp(ic))=newpar(pa(ic),xp(ic))+r(ic).*xy; % 2nd
  if xp(ic)<npar % crossover when last variable not selected
    newpar(keep+ix(ic),:)=[newpar(keep+ix(ic),1:xp(ic)) newpar(keep+ix(ic)+1,xp(ic)+1:npar)];
    newpar(keep+ix(ic)+1,:)=[newpar(keep+ix(ic)+1,1:xp(ic)) newpar(keep+ix(ic),xp(ic)+1:npar)];
  endif
endfor

## Generate Next Generation's par Matrix

par = newpar;

## Mutation Routine

mrow=sort(ceil(rand(1,nmut)*(popsize-1))+1);
mcol=ceil(rand(1,nmut)*Nt);

for ii=1:nmut
  ## Formatting Random Numbers for Fourier Series coefficients
  par(mrow(ii),mcol(ii))=(constantshi-constantslo)*rand+constantslo; 
  
  ## Formatting Amplitude Values
  for col=(npar/6):(npar/6):npar
    if mcol(ii) == col
      if par(mrow(ii),col) < 0
        par(mrow(ii),col)=(-1)*par(mrow(ii),col); % Amplitude scaling must be +ve
      endif   
      ## Formating Random Numbers for Pertubation Height
      par(mrow(ii),mcol(ii))=((heighthi-heightlow)*par(mrow(ii),mcol(ii))+heightlow); 
    endif
  endfor 
endfor

## Generating updated system log

for a=1:1:popsize
  for b=1:1:npar
    log(((popsize*iga)+a),b)=par(a,b);
  endfor
endfor

filename = ['Gen_' int2str(iga)];
save(filename, 'log', 'iga', 'par', 'fittest', 'minc', 'meanc')

endfunction
